@@ -52,7 +52,6 @@ class Elaborator:
         self.tops = tops
         self.ctx = ctx
         self.stack: List[ElabStackEntry] = list()
-        self.modules: Dict[int, Module] = dict()
 
     def elaborate_tops(self) -> List[Elaboratable]:
         """Elaborate our top nodes"""
@@ -89,12 +88,9 @@ class Elaborator:
         `elaborate_module_base` instead.
         """
 
-        if id(module) in self.modules:  # Check our cache
-            return module  # Already done!
-
-        if not module.name:
-            msg = f"Anonymous Module {module} cannot be elaborated (did you forget to name it?)"
-            self.fail(msg)
+        # Check if this has already been elaborated
+        if module._elaborated is not None:
+            return module._elaborated
 
         self.stack.append(module)
 
@@ -113,8 +109,7 @@ class Elaborator:
         # Run the pass-specific `elaborate_module`
         result = self.elaborate_module(module)
 
-        # Store a reference to the now-elaborated Module in our cache, and return it
-        self.modules[id(module)] = result
+        # Pop the hierarchy-stack and return it
         self.stack.pop()
         return result
 
@@ -185,11 +180,21 @@ class Elaborator:
 
     def fail(self, msg: str):
         """Error helper, adding stack and state info to an error"""
+        lines = self.format_hier_path()
+        msg = "Elaboration Error at hierarchical path: \n" + "".join(lines) + msg
+        # This also serves as a very helpful place for a debugger breakpoint.
+        raise RuntimeError(msg)
+
+    def format_hier_path(self) -> List[str]:
+        """Create a list of lines describing the current hierarchy path
+        Generally intended for debug use, especially in error messages."""
 
         lines = []
         for s in self.stack:
             # Format: `  Module          MyModule      `
-            line = "  " + type(s).__name__.ljust(14) + str(s.name).ljust(28)
+            s_name = s.name or ""
+            # Filter out None-valued names, common during elaboration errors.
+            line = "  " + type(s).__name__.ljust(14) + s_name.ljust(28)
 
             source_info = getattr(s, "_source_info", None)
             if source_info is not None:
@@ -213,7 +218,4 @@ class Elaborator:
 
             line += "\n"
             lines.append(line)
-
-        msg = "Elaboration Error at hierarchical path: \n" + "".join(lines) + msg
-        # This also serves as a very helpful place for a debugger breakpoint.
-        raise RuntimeError(msg)
+        return lines
