@@ -15,6 +15,11 @@ from ...instantiable import Instantiable
 from .base import Elaborator
 
 
+# Cache GeneratorCalls to their (Module) results
+# (Yes, GeneratorCalls can be hashed.)
+THE_GENERATOR_CALL_CACHE: Dict[GeneratorCall, Module] = dict()
+
+
 class GeneratorElaborator(Elaborator):
     """
     # Generator Elaborator
@@ -32,12 +37,6 @@ class GeneratorElaborator(Elaborator):
     and `GeneratorElaborator`'s special-ish case is left to over-ride it.
     """
 
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        # Cache GeneratorCalls to their (Module) results
-        # (Yes, GeneratorCalls can be hashed.)
-        self.generator_calls: Dict[GeneratorCall, Module] = dict()
-
     def elaborate_tops(self) -> List[Module]:
         """Elaborate our top nodes"""
         if not isinstance(self.tops, List):
@@ -59,13 +58,18 @@ class GeneratorElaborator(Elaborator):
     def elaborate_generator_call(self, call: GeneratorCall) -> Module:
         """Elaborate Generator-function-call `call`. Returns the generated Module."""
 
-        # Check our cache, see if we've already generated its module.
-        if call in self.generator_calls:  # Already done!
-            # Give the `call` a reference to its result.
-            # Note this *has not* necessarily already happened, as the `self.generator_calls` key may be an equally-valued (but distinct) `GeneratorCall`.
-            result = self.generator_calls[call]
-            call.result = result
-            return result
+        # First and foremost - caching.
+        # See if we've already run this generator-parameters combo.
+        # We store this both in the module-scope cache, and on the Call itself.
+        # If the two are different Modules... not sure what would cause that, but fail.
+        global THE_GENERATOR_CALL_CACHE
+        cached_result = THE_GENERATOR_CALL_CACHE.get(call, None)
+        if cached_result is not None:
+            if call.result is not None and call.result is not cached_result:
+                msg = f"GeneratorCall {call} has two different results: {call.result} and {cached_result}"
+                self.fail(msg)
+            call.result = cached_result
+            return call.result
 
         # Add both the `Call` and `Generator` to our stack.
         self.stack.append(call)
@@ -85,31 +89,37 @@ class GeneratorElaborator(Elaborator):
         except Exception as e:
             self.fail(f"{call.gen} raised an exception: \n{e}")
 
-        # Type-check the result
-        # Generators may return other (potentially nested) generator-calls; unwind any of them
-        while isinstance(m, GeneratorCall):
-            # Note this should hit Python's recursive stack-check if it doesn't terminate
-            m = self.elaborate_generator_call(m)
+        if isinstance(m, Module):
+            # Give the result a reference back to the generating `Call`
+            m._generated_by = call
+
+            # And elaborate the module
+            m = self.elaborate_module_base(m)  # Note the `_base` here!
+
+            # If the Module that comes back is anonymous, start by giving it a name equal to the Generator's
+            if m.name is None:
+                m.name = call.gen.func.__name__
+
+            # Then add a unique suffix per its parameter-values
+            # Note this part may require that `m` has been through elaboration above!
+            if not isinstance(call.params, HasNoParams):
+                m.name += "(" + _unique_name(call.params) + ")"
+
+        # Generators may return other (potentially nested) generator-calls; recursively unwind any of them
+        # Note this should hit Python's recursive stack-check if it doesn't terminate
+        elif isinstance(m, GeneratorCall):
+            m._generated_by = call
+            m: Module = self.elaborate_generator_call(m)
+
+        # Type-check the result.
         # Ultimately they've gotta resolve to Modules, or they fail.
-        if not isinstance(m, Module):
+        else:
             msg = f"Generator {call.gen.func.__name__} returned {type(m)}, must return Module."
             self.fail(msg)
 
-        # Give the GeneratorCall a reference to its result, and store it in our local dict
+        # Store the result in our cache, and on the Call.
         call.result = m
-        self.generator_calls[call] = m
-
-        # Create a unique name
-        # If the Module that comes back is anonymous, start by giving it a name equal to the Generator's
-        if m.name is None:
-            m.name = call.gen.func.__name__
-
-        # Then add a unique suffix per its parameter-values
-        if not isinstance(call.params, HasNoParams):
-            m.name += "(" + _unique_name(call.params) + ")"
-
-        # And elaborate the module
-        m = self.elaborate_module_base(m)  # Note the `_base` here!
+        THE_GENERATOR_CALL_CACHE[call] = m
 
         # Pop both the `Call` and `Generator` off the stack
         self.stack.pop()
