@@ -8,7 +8,7 @@ This module primarily defines:
 """
 
 from inspect import isclass
-from typing import Any, Optional, Union, List
+from typing import Any, Optional, Union, List, Dict
 
 # Local imports
 from .source_info import source_info, SourceInfo
@@ -57,7 +57,7 @@ class Module:
     Parametric hardware is produced through Hdl21's `generator` facility, which defines python functions which create and return `Module`s.
     """
 
-    def __init__(self, *, name: Optional[str] = None):
+    def __init__(self, name: Optional[str] = None):
         if name is not None and not isinstance(name, str):
             # Something wrong with this `name`.
             # Most common case: confusion with the `module` decorator.
@@ -76,8 +76,21 @@ class Module:
         self.bundles = dict()
         self.namespace = dict()  # Combination of all these
 
-        self._source_info: Optional[SourceInfo] = source_info(get_pymodule=True)
+        # Elaborated version of this module.
+        # Set at the end of elaboration.
+        # For most modules this will be `self`.
+        self._elaborated: Optional[Module] = None
+
+        # IOs as captured before bundle-flattening.
+        # Bundle-valued ports are flattened into `ports` and removed from `bundles`, but need to be kept *somewhere* afterwards.
+        # Set to `None` initially to indicate that it hasn't been set yet.
+        self._pre_flattening_io: Optional[Dict[str, "Connectable"]] = None
+
+        # The source `GeneratorCall`, for generated Modules.
+        self._generated_by: Optional["GeneratorCall"] = None
+
         self._importpath = None  # Optional field set by importers
+        self._source_info: Optional[SourceInfo] = source_info(get_pymodule=True)
         self._initialized = True
 
     """
@@ -195,6 +208,14 @@ class Module:
             return f"Module(name={self.name})"
         return f"Module(_anon_)"
 
+    def __eq__(self, other) -> bool:
+        # Identity is equality
+        return id(self) == id(other)
+
+    def __hash__(self) -> bool:
+        # Identity is equality
+        return hash(id(self))
+
 
 def module(cls: type) -> Module:
     """
@@ -285,6 +306,9 @@ def _add(module: Module, val: ModuleAttr) -> ModuleAttr:
     Primarily sort `val` into one of our type-based containers.
     Layers above `_add` must ensure that `val` has its `name` attribute before calling this method."""
 
+    if module._elaborated is not None:
+        raise RuntimeError(f"Cannot add {val} to {module} after elaboration.")
+
     if isinstance(val, Signal):
         module.namespace[val.name] = val
         if val.vis == Visibility.PORT:
