@@ -71,10 +71,10 @@ from pydantic.dataclasses import dataclass
 # Local imports
 from .default import Default
 from .call import param_call
-from .params import paramclass, Param, isparamclass, NoParams
+from .params import paramclass, Param, isparamclass, NoParams, _unique_name
 from .signal import Port, Signal, Visibility
 from .instance import calls_instantiate
-from .prefix import Prefix, Prefixed
+from .scalar import Scalar
 
 
 class PrimitiveType(Enum):
@@ -117,12 +117,21 @@ class Primitive:
 
     @property
     def Params(self) -> Type:
+        """Type-style alias for the parameter-type."""
         return self.paramtype
 
     @property
     def ports(self) -> Dict[str, Signal]:
         return {p.name: p for p in self.port_list}
 
+    def __eq__(self, other) -> bool:
+        # Identity is equality
+        return id(self) == id(other)
+
+    def __hash__(self) -> bool:
+        # Identity is equality
+        return hash(id(self))
+
 
 @calls_instantiate
 @dataclass
@@ -140,13 +149,32 @@ class PrimitiveCall:
             msg = f"Invalid parameters {self.params} for Primitive {self.prim}. Must be {self.prim.paramtype}"
             raise TypeError(msg)
 
+    @property
+    def name(self) -> str:
+        return self.prim.name + "(" + _unique_name(self.params) + ")"
+
     @property
     def ports(self) -> dict:
         return self.prim.ports
 
+    def __eq__(self, other) -> bool:
+        """Call equality requires:
+        * *Identity* between prims, and
+        * *Equality* between parameter-values."""
+        return self.prim is other.prim and self.params == other.params
+
+    def __hash__(self):
+        """Generator-Call hashing, consistent with `__eq__` above, uses:
+        * *Identity* of its prim, and
+        * *Value* of its parameters.
+        The two are joined for hashing as a two-element tuple."""
+        return hash((id(self.prim), self.params))
+
 
 @dataclass
 class PrimLibEntry:
+    """# Entry in the Primitive Library"""
+
     prim: Primitive
     aliases: List[str]
 
@@ -201,8 +229,8 @@ class MosVth(Enum):
 class MosParams:
     """MOS Transistor Parameters"""
 
-    w = Param(dtype=Optional[Prefixed], desc="Width in resolution units", default=None)
-    l = Param(dtype=Optional[Prefixed], desc="Length in resolution units", default=None)
+    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
+    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
     npar = Param(dtype=int, desc="Number of parallel fingers", default=1)
     tp = Param(dtype=MosType, desc="MosType (Nmos/ Pmos)", default=MosType.NMOS)
     vth = Param(dtype=MosVth, desc="Threshold voltage specifier", default=MosVth.STD)
@@ -240,14 +268,18 @@ Mos = _add(
 )
 
 
-def Nmos(params: MosParams) -> Primitive:
+def Nmos(arg: Any = Default, **kwargs) -> Primitive:
     """Nmos Constructor. A thin wrapper around `hdl21.primitives.Mos`"""
-    return Mos(replace(params, tp=MosType.NMOS))
+    mos = Mos(arg, **kwargs)
+    mos.params = replace(mos.params, tp=MosType.NMOS)
+    return mos
 
 
-def Pmos(params: MosParams) -> Primitive:
+def Pmos(arg: Any = Default, **kwargs) -> Primitive:
     """Pmos Constructor. A thin wrapper around `hdl21.primitives.Mos`"""
-    return Mos(replace(params, tp=MosType.PMOS))
+    mos = Mos(arg, **kwargs)
+    mos.params = replace(mos.params, tp=MosType.PMOS)
+    return mos
 
 
 """ 
@@ -262,7 +294,7 @@ ThreeTerminalPorts = [Port(name="p"), Port(name="n"), Port(name="b")]
 
 @paramclass
 class ResistorParams:
-    r = Param(dtype=float, desc="Resistance (ohms)")
+    r = Param(dtype=Scalar, desc="Resistance (ohms)")
 
 
 _add(
@@ -279,8 +311,8 @@ _add(
 
 @paramclass
 class PhysicalResistorParams:
-    w = Param(dtype=Optional[Prefixed], desc="Width in resolution units", default=None)
-    l = Param(dtype=Optional[Prefixed], desc="Length in resolution units", default=None)
+    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
+    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
     model = Param(dtype=Optional[str], desc="Model (Name)", default=None)
 
 
@@ -310,7 +342,7 @@ _add(
 
 @paramclass
 class IdealCapacitorParams:
-    c = Param(dtype=float, desc="Capacitance (F)")
+    c = Param(dtype=Scalar, desc="Capacitance (F)")
 
 
 _add(
@@ -327,7 +359,7 @@ _add(
 
 @paramclass
 class PhysicalCapacitorParams:
-    c = Param(dtype=float, desc="Capacitance (F)")
+    c = Param(dtype=Scalar, desc="Capacitance (F)")
 
 
 _add(
@@ -356,7 +388,7 @@ _add(
 
 @paramclass
 class IdealInductorParams:
-    l = Param(dtype=float, desc="Inductance (H)")
+    l = Param(dtype=Scalar, desc="Inductance (H)")
 
 
 _add(
@@ -373,7 +405,7 @@ _add(
 
 @paramclass
 class PhysicalInductorParams:
-    l = Param(dtype=float, desc="Inductance (H)")
+    l = Param(dtype=Scalar, desc="Inductance (H)")
 
 
 _add(
@@ -403,8 +435,8 @@ _add(
 @paramclass
 class PhysicalShortParams:
     layer = Param(dtype=Optional[Union[int, str]], desc="Metal layer", default=None)
-    w = Param(dtype=Optional[Prefixed], desc="Width in resolution units", default=None)
-    l = Param(dtype=Optional[Prefixed], desc="Length in resolution units", default=None)
+    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
+    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
 
 
 _add(
@@ -428,8 +460,8 @@ Sources
 class DcVoltageSourceParams:
     """`DcVoltageSource` Parameters"""
 
-    dc = Param(dtype=Optional[Prefixed], default=0, desc="DC Value (V)")
-    ac = Param(dtype=Optional[Prefixed], default=None, desc="AC Amplitude (V)")
+    dc = Param(dtype=Optional[Scalar], default=0, desc="DC Value (V)")
+    ac = Param(dtype=Optional[Scalar], default=None, desc="AC Amplitude (V)")
 
 
 _add(
@@ -452,13 +484,13 @@ _add(
 class PulseVoltageSourceParams:
     """`PulseVoltageSource` Parameters"""
 
-    delay = Param(dtype=Optional[Prefixed], default=None, desc="Time Delay (s)")
-    v1 = Param(dtype=Optional[Prefixed], default=None, desc="One Value (V)")
-    v2 = Param(dtype=Optional[Prefixed], default=None, desc="Zero Value (V)")
-    period = Param(dtype=Optional[Prefixed], default=None, desc="Period (s)")
-    rise = Param(dtype=Optional[Prefixed], default=None, desc="Rise time (s)")
-    fall = Param(dtype=Optional[Prefixed], default=None, desc="Fall time (s)")
-    width = Param(dtype=Optional[Prefixed], default=None, desc="Pulse width (s)")
+    delay = Param(dtype=Optional[Scalar], default=None, desc="Time Delay (s)")
+    v1 = Param(dtype=Optional[Scalar], default=None, desc="One Value (V)")
+    v2 = Param(dtype=Optional[Scalar], default=None, desc="Zero Value (V)")
+    period = Param(dtype=Optional[Scalar], default=None, desc="Period (s)")
+    rise = Param(dtype=Optional[Scalar], default=None, desc="Rise time (s)")
+    fall = Param(dtype=Optional[Scalar], default=None, desc="Fall time (s)")
+    width = Param(dtype=Optional[Scalar], default=None, desc="Pulse width (s)")
 
 
 _add(
@@ -473,9 +505,32 @@ _add(
 )
 
 
+@paramclass
+class SineVoltageSourceParams:
+    """`SineVoltageSource` Parameters"""
+
+    voff = Param(dtype=Optional[Scalar], default=None, desc="Offset (V)")
+    vamp = Param(dtype=Optional[Scalar], default=None, desc="Amplitude (V)")
+    freq = Param(dtype=Optional[Scalar], default=None, desc="Frequency (Hz)")
+    td = Param(dtype=Optional[Scalar], default=None, desc="Delay (s)")
+    phase = Param(dtype=Optional[Scalar], default=None, desc="Phase at td (degrees)")
+
+
+_add(
+    prim=Primitive(
+        name="SineVoltageSource",
+        desc="Sine Voltage Source",
+        port_list=copy.deepcopy(PassivePorts),
+        paramtype=SineVoltageSourceParams,
+        primtype=PrimitiveType.IDEAL,
+    ),
+    aliases=["Vsin"],
+)
+
+
 @paramclass
 class CurrentSourceParams:
-    dc = Param(dtype=Optional[Prefixed], default=0, desc="DC Value (A)")
+    dc = Param(dtype=Optional[Scalar], default=0, desc="DC Value (A)")
 
 
 _add(
@@ -497,7 +552,7 @@ Controlled Sources
 
 @paramclass
 class ControlledSourceParams:
-    gain = Param(dtype=Prefixed, default=1 * Prefix.UNIT, desc="Gain in SI Units")
+    gain = Param(dtype=Scalar, default=1, desc="Gain in SI Units")
 
 
 # Controlled Sources Port List
@@ -565,8 +620,8 @@ class BipolarType(Enum):
 class BipolarParams:
     """Bipolar Transistor Parameters"""
 
-    w = Param(dtype=Optional[Prefixed], desc="Width in resolution units", default=None)
-    l = Param(dtype=Optional[Prefixed], desc="Length in resolution units", default=None)
+    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
+    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
     tp = Param(
         dtype=BipolarType, desc="Bipolar Type (NPN/ PNP)", default=BipolarType.NPN
     )
@@ -593,14 +648,18 @@ Bipolar = _add(
 )
 
 
-def Npn(params: BipolarParams) -> Primitive:
+def Npn(arg: Any = Default, **kwargs) -> Primitive:
     """Npn Constructor. A thin wrapper around `hdl21.primitives.Bipolar`"""
-    return Bipolar(replace(params, tp=BipolarType.NPN))
+    bip = Bipolar(arg, **kwargs)
+    bip.params = replace(bip.params, tp=BipolarType.NPN)
+    return bip
 
 
-def Pnp(params: BipolarParams) -> Primitive:
+def Pnp(arg: Any = Default, **kwargs) -> Primitive:
     """Pnp Constructor. A thin wrapper around `hdl21.primitives.Bipolar`"""
-    return Bipolar(replace(params, tp=BipolarType.PNP))
+    bip = Bipolar(arg, **kwargs)
+    bip.params = replace(bip.params, tp=BipolarType.PNP)
+    return bip
 
 
 """ 
@@ -610,8 +669,8 @@ Diodes
 
 @paramclass
 class DiodeParams:
-    w = Param(dtype=Optional[Prefixed], desc="Width in resolution units", default=None)
-    l = Param(dtype=Optional[Prefixed], desc="Length in resolution units", default=None)
+    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
+    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
     model = Param(dtype=Optional[str], desc="Model (Name)", default=None)
     # FIXME: whether to include `model`
 
