@@ -413,26 +413,61 @@ def test_slice_module1():
     h.elaborate(P)
 
 
-def test_module_as_param():
-    """Test using a `Module` as a parameter-value"""
+def test_instantiable_as_param():
+    """Test using each Instantiable type as a parameter-value."""
 
     @h.paramclass
-    class HasModuleParam:
-        m = h.Param(dtype=h.Module, desc="A `Module` provided as a parameter")
-        e = h.Param(
-            dtype=h.ExternalModule, desc="An `ExternalModule` provided as a parameter"
-        )
+    class Params:
+        m = h.Param(dtype=h.Module, desc="A `Module`")
+        e = h.Param(dtype=h.ExternalModule, desc="An `ExternalModule`")
+        ec = h.Param(dtype=h.ExternalModuleCall, desc="An `ExternalModuleCall`")
+        # FIXME: `Generator` is the one exception here. Maybe it can work, some day.
+        # g = h.Param(dtype=h.Generator, desc="An `Generator`")
+        gc = h.Param(dtype=h.GeneratorCall, desc="An `GeneratorCall`")
+        p = h.Param(dtype=h.Primitive, desc="An `Primitive`")
+        pc = h.Param(dtype=h.PrimitiveCall, desc="An `PrimitiveCall`")
+        i = h.Param(dtype=h.Instantiable, desc="An `Instantiable`")
 
     @h.generator
-    def UsesModuleParam(params: HasModuleParam) -> h.Module:
-        return params.m  # Returns the Module unmodified
+    def UsesThemParams(params: Params) -> h.Module:
+        @h.module
+        class M:
+            x = h.Signal()
+
+            m = params.m(x=x)
+            e = params.e()(x=x)
+            ec = params.ec(x=x)
+            # g = params.g()(x=x)
+            gc = params.gc(x=x)
+            p = params.p()(d=x, g=x, s=x, b=x)
+            pc = params.pc(d=x, g=x, s=x, b=x)
+            i = params.i(x=x)
+
+        return M
 
     Mod = h.Module(name="Mod")
-    Emod = h.ExternalModule(name="Emod", port_list=[])
-    p = HasModuleParam(m=Mod, e=Emod)
-    m = UsesModuleParam(p)
+    Mod.x = h.Port()
+    Emod = h.ExternalModule(name="Emod", port_list=[h.Port(name="x")])
+
+    @h.generator
+    def Gen(_: h.HasNoParams) -> h.Module:
+        m = h.Module()
+        m.x = h.Port()
+        return m
+
+    p = Params(
+        m=Mod,
+        e=Emod,
+        ec=Emod(),
+        # g=Gen,
+        gc=Gen(),
+        p=h.Mos,
+        pc=h.Mos(),
+        i=Mod,
+    )
+    m = UsesThemParams(p)
     m = h.elaborate(m)
-    assert m == Mod
+    # assert m == Mod
 
 
 def test_instance_mult():
@@ -706,20 +741,26 @@ def test_orphanage4():
 def test_wrong_decorator():
     """Mistake `Module` for `module` and vice versa"""
 
-    with pytest.raises(TypeError):
+    with pytest.raises(TypeError) as e:
 
         @h.Module  # Bad!
         class M:
             ...
 
-    with pytest.raises(TypeError):
+    assert "Did you mean to use the `module` decorator?" in str(e)
+
+    with pytest.raises(TypeError) as e:
 
         h.Module(2)  # Bad!
 
-    with pytest.raises(TypeError):
+    assert "Invalid Module name" in str(e)
+
+    with pytest.raises(TypeError) as e:
         ok = h.Module("ok")  # OK
         h.Module(ok)  # Bad!
 
+    assert "Invalid Module name" in str(e)
+
 
 def test_elab_noconn():
     """Initial test of elaborating a `NoConn`"""
@@ -738,6 +779,28 @@ def test_elab_noconn():
     assert len(HasNoConn.signals) == 1
 
 
+def test_elab_noconn2():
+    """Slightly more elaborate test of elaborating a `NoConn`"""
+
+    @h.module
+    class M3:  # Layer 3: just a Port
+        p = h.Port()
+
+    @h.module
+    class M2:  # Layer 2: instantiate & connect M3
+        p = h.Port()
+        m3 = M3(p=p)
+
+    @h.module
+    class M1:  # Layer 1: connect a `NoConn` to all that
+        p = h.NoConn()
+        m2 = M2(p=p)
+
+    h.elaborate(M1)
+
+    assert len(M1.signals) == 1
+
+
 def test_bad_noconn():
     """Test that a doubly-connected `NoConn` should fail"""
 
@@ -994,7 +1057,7 @@ def test_bad_conns():
 
     with pytest.raises(RuntimeError) as e:
         h.elaborate(O)
-    assert "Connections to invalid Ports" in str(e.value)
+    assert "Connection to non-existent Port" in str(e.value)
 
     @h.module
     class O:
@@ -1004,6 +1067,7 @@ def test_bad_conns():
     with pytest.raises(RuntimeError) as e:
         h.elaborate(O)
     assert "Missing connection to" in str(e.value)
+    assert "Connection to non-existent Port" in str(e.value)
 
     @h.module
     class O:
@@ -1012,7 +1076,7 @@ def test_bad_conns():
 
     with pytest.raises(RuntimeError) as e:
         h.elaborate(O)
-    assert "Connection to invalid Port" in str(e.value)
+    assert "Connection to non-existent Port" in str(e.value)
 
     @h.module
     class O:
@@ -1081,7 +1145,6 @@ def test_pair1():
     assert len(O.instbundles) == 0
 
 
-@pytest.mark.xfail(reason="#33 https://github.com/dan-fritchman/Hdl21/issues/33")
 def test_noconn_types():
     """Test connecting `NoConn`s to a variety of port-types."""
 
@@ -1096,8 +1159,18 @@ def test_noconn_types():
     class Outer:
         i = Inner(s=h.NoConn(), b=h.NoConn(), d=h.NoConn())
 
+    # Most of the test is here, just seeing that this can elaborate
     h.elaborate(Outer)
 
+    assert len(Inner.signals) == 0
+    assert len(Outer.ports) == 0
+    assert sorted(Inner.ports.keys()) == ["b", "d_n", "d_p", "s"]
+    assert sorted(Outer.signals.keys()) == ["i_b", "i_d_n", "i_d_p", "i_s"]
+    assert Outer.i.conns["b"] is Outer.i_b
+    assert Outer.i.conns["d_n"] is Outer.i_d_n
+    assert Outer.i.conns["d_p"] is Outer.i_d_p
+    assert Outer.i.conns["s"] is Outer.i_s
+
 
 def test_deep_hierarchy():
     """Test a deep hierarchy, without much content."""
@@ -1117,10 +1190,7 @@ def test_deep_hierarchy():
             M1 = m
         prev = m
 
-    # Elaborate the final, highest-level module.
-    h.elaborate(m)
-
-    # Screw up a connection, and check we get a RuntimeError.
+    # Screw up a connection, and check that we get an elaboration-time error when elaborating the top-level module.
     M1.i.not_a_real_port = M1.p
     with pytest.raises(RuntimeError):
         h.elaborate(m)
@@ -1187,3 +1257,65 @@ def test_param_calls():
         R("not_a_ResistorParams_as_positional_arg")
     with pytest.raises(RuntimeError):
         R(R.Params(r=11), z=12)
+
+
+def test_bad_generators():
+    """Test generator functions with bad call-signatures."""
+
+    @h.paramclass
+    class P:
+        ...
+
+    @h.generator  # A good, working generator!
+    def good(p: P) -> h.Module:
+        return h.Module()
+
+    with pytest.raises(RuntimeError):
+        # Bad 1st argument type
+        @h.generator
+        def bad(p: TabError) -> h.Module:
+            return h.Module()
+
+    with pytest.raises(RuntimeError):
+        # Bad 2nd argument type
+        @h.generator
+        def bad(p: P, ctx: TabError) -> h.Module:
+            return h.Module()
+
+    with pytest.raises(RuntimeError):
+        # Extra arg
+        @h.generator
+        def bad(p: P, ctx: h.Context, something_else: int) -> h.Module:
+            return h.Module()
+
+    with pytest.raises(RuntimeError):
+        # Bad return type
+        @h.generator
+        def bad(p: P) -> TabError:
+            return h.Module()
+
+
+def test_multi_portref_conns():
+    """Test making several PortRef-based connections with a single assignment line."""
+
+    @h.module
+    class I:  # Inner Module
+        a, b = h.Ports(2)
+
+    @h.module
+    class O:  # Outer, instantiating Module
+        i1 = I()
+        i2 = I()
+        i3 = I()
+
+        # The several connections
+        i1.a = i2.a = i3.a
+
+    # And do so again outside the class body
+    O.i1.b = O.i2.b = O.i3.b
+
+    h.elaborate(O)
+
+    assert list(O.signals.keys()) == ["i3_a", "i3_b"]
+    assert O.i1.conns["a"] is O.i2.conns["a"] is O.i3.conns["a"] is O.i3_a
+    assert O.i1.conns["b"] is O.i2.conns["b"] is O.i3.conns["b"] is O.i3_b
